// framework.hxx
//
// Mical, a regular grammar inference library
// Copyright (C) 2003 Maxime Rey
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MICAL_FRAMEWORK_HXX_
#define MICAL_FRAMEWORK_HXX_

#include <mical/framework/framework.hh>

#include <mical/framework/summary.hh>
#include <mical/internal/summary.hh>
#include <mical/samples/services/summary.hh>
#include <mical/samples/tools/load_sample.hh>

#include <mical/framework/determin.hh>


namespace mical {

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  FrameWork()
  {
    view_ = 0;
    auto_.create();
  }
   
 
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  FrameWork(const FrameWork& other)
  {
    auto_ = other.auto_;
    auto_.emancipate();

    partition_ = other.partition_;

    view_ = new view_t(auto_, partition_);   
    view_->update();

    sample_ = other.sample_;
  }

  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  FrameWork(char* name)
  {
    load_presentation(sample_, name);

    auto_.create();

    typename automaton_t::monoid_t::alphabet_t alphabet;
    give_alphabet(auto_, sample_, alphabet);
    
    auto_.series().monoid().alphabet() = alphabet;
    
    PROXY3(create_automaton, search_kind_t(), auto_, sample_);
    
    for (typename automaton_t::state_iterator i = auto_.states().begin();
	 i != auto_.states().end(); i++)
      partition_.insert(*i);
    
    view_ = new view_t(auto_, partition_);
  }

  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  FrameWork(const sample_t& sample)
  {
    sample_ = sample;
    
    auto_.create();

    typename automaton_t::monoid_t::alphabet_t alphabet;
    give_alphabet(auto_, sample_, alphabet);
    
    auto_.series().monoid().alphabet() = alphabet;

    PROXY3(create_automaton, search_kind_t(), auto_, sample_);
    
    for (typename automaton_t::state_iterator i = auto_.states().begin();
	 i != auto_.states().end(); i++)
      partition_.insert(*i);
    
    view_ = new view_t(auto_, partition_);
  }

  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  FrameWork(const automaton_t& automaton, const sample_t& sample)
  {
    sample_ = sample;

    auto_ = automaton;
    auto_.emancipate();

    typename automaton_t::monoid_t::alphabet_t alphabet;
    give_alphabet(auto_, sample_, alphabet);
    
    auto_.series().monoid().alphabet() = alphabet;
    
    for (typename automaton_t::state_iterator i = auto_.states().begin();
	 i != auto_.states().end(); i++)
      partition_.insert(*i);
    
    view_ = new view_t(auto_, partition_);    
  }
    

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  init()
  {   
    partition_.clear();
    
    PROXY2(init, inference_kind_t(), sample_);
    
    for (typename automaton_t::state_iterator i = auto_.states().begin();
	 i != auto_.states().end(); i++)
      partition_.insert(*i);
    
    view_->update();
  }

  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  fusion(hstate_t lhs, hstate_t rhs)
  {
    if (lhs == rhs)
      return;

    PROXY5(fusion, search_kind_t(), *view_, partition_, lhs, rhs);
    
    view_->update();
  }
   
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  fusion(state_iterator& lhs, state_iterator& rhs)
  {
    hstate_t min;
    if (*lhs < *rhs)
      min = *lhs;
    else
      min = *rhs;

    if (*lhs == *rhs)
      return;

    PROXY5(fusion, search_kind_t(), *view_, partition_, *lhs, *rhs);
    
    view_->update();

    lhs = find(view_->states().begin(), view_->states().end(), min);
    rhs = find(view_->states().begin(), view_->states().end(), min);
  }


  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  fission(hstate_t lhs, hstate_t rhs)
  {
    if (lhs == rhs)
      return;

    PROXY5(fission, search_kind_t(), *view_, partition_, lhs, rhs);
    
    view_->update();
  }
  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  fission(state_iterator& lhs, state_iterator& rhs)
  {
    hstate_t lhs_state = *lhs;
    hstate_t rhs_state = *rhs;

    if (*lhs == *rhs)
      return;
    

    PROXY5(fission, search_kind_t(), *view_, partition_, *lhs, *rhs);
    
    view_->update();
    
    lhs = find(view_->states().begin(), view_->states().end(), lhs_state);
    rhs = find(view_->states().begin(), view_->states().end(), rhs_state);
  }

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  fission(hstate_t lhs, const block_t& rhs)
  {
    if (!rhs.size())
      return;

    for (typename block_t::iterator i = rhs.begin(); i != rhs.end(); i++)
      if (lhs != *i)
	PROXY5(fission, search_kind_t(), *view_, partition_, lhs, *i);

    for (typename block_t::iterator i = rhs.begin(); i != rhs.end(); i++)
      if (i != rhs.begin())
	PROXY5(fusion, search_kind_t(), *view_, partition_, *rhs.begin(), *i);
    
    view_->update();
  }


  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_universal() const
  {
    return (view_->states().size() == 1);
  }


  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_null() const
  {
    return (partition_.size() == auto_.states().size());
  }

    

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_compatible() const
  {
    return mical::is_compatible(*view_, sample_);
  }
  

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_lattice_elt() const
  {
    return is_sc_sample(*view_, sample_);
  }

  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_lattice_elt(const automaton_t& automaton) const
  {
    return is_sc_sample(automaton, sample_);
  }


  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  bool 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  is_borderset_elt()
  {
    bool res = true;

    if (!is_compatible())
      return false;

    for (typename view_t::state_iterator i = view_->states().begin();
	 i != view_->states().end(); i++)
      if (res)
	for (typename view_t::state_iterator j = i; 
	     j != view_->states().end(); j++)
	  if ((j != i) && (res))
	    {
	      hstate_t a = *i, b = *j;

	      block_t svg = *partition_.find_elt(b);

	      fusion(a, b);

	      if (is_compatible())
		res = false;

	      fission(a, svg);

	      i = view_->update(a);
	      j = view_->update(b);
	    }
    
    return res;
  }
  

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  determin(hstate_t state)
  {
    PROXY4(determin, search_kind_t(), *view_, state, partition_);
  }


  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  template <class Any>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  apply(Any func)
  {
    func(*this);
  }
  
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  typename Implem_::view_t*
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  view() const
  {
    return view_;
  }

  // to disapear
  /////////////////////////////////////////////////////
  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  typename Implem_::partition_t&
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  partition()
  {
    return partition_;
  }
  /////////////////////////////////////////////////////

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  typename Implem_::automaton_t 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  generate() const
  {
    automaton_t res;
    res.create();

    mical::generate(*view_, res);

    return res;
  }

  template <typename SearchKind_, 
	    typename InferenceKind_,
	    typename Implem_>
  template <class Any>
  void 
  FrameWork<SearchKind_, InferenceKind_, Implem_>::
  next_step(Any func)
  {
    PROXY2(next_step, inference_kind_t(), sample_);
    
    apply(func);
  }


} // mical

#endif // MICAL_FRAMEWORK_HXX_
