// view_lattice_elt.hxx
//
// Mical, a regular grammar inference library
// Copyright (C) 2003 Maxime Rey
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MICAL_VIEW_LATTICE_ELT_HXX_
#define MICAL_VIEW_LATTICE_ELT_HXX_

#include <mical/framework/viewer/view_lattice_elt.hh>

#include <algorithm>
#include <vector>

namespace vcsn {
  
  using std::vector;

  template<typename T> 
  Element<ViewLattice, T>::
  Element()
  {
    auto_ = 0;
    partition_ = 0;
  }

  template<typename T> 
  Element<ViewLattice, T>::
  Element(automaton_t& a, partition_t& partition)
    : auto_(&a), partition_(&partition)
  {
    update();
  }

  template<typename T> 
  void
  Element<ViewLattice, T>::
  emancipate()
  {}

  template<typename T> 
  void 
  Element<ViewLattice, T>::
  create()
  {}

  template<typename T> 
  hstate_t
  Element<ViewLattice, T>::get_min_block(hstate_t state) const
  {
    for (typename partition_t::block_iterator i = partition_->begin();
	 i != partition_->end(); i++)
      {
	hstate_t min = *i->begin();
	bool flag = false;

	for (typename block_t::iterator j = i->begin(); j != i->end(); j++)
	  {
	    if (*j == state)
	      flag = true; 
	    if (*j < min)
	      min = *j;
	  }
	if (flag)
	  return min;
      }
    return -1;
  }
  
  template<typename T> 
  typename T::partition_t::block_t
  Element<ViewLattice, T>::get_block(hstate_t state) const
  {
    for (typename partition_t::block_iterator i = partition_->begin();
	 i != partition_->end(); i++)
      {
	bool flag = false;
	for (typename block_t::iterator j = i->begin(); j != i->end(); j++)
	  if (*j == state)
	    flag = true;
	
	if (flag)
	  return *i;
      }
    
    return block_t();
  }


  template<typename T> 
  void 
  Element<ViewLattice, T>::update()
  {
    ////////////////////////////////////////////////////////////////////////////
    states_.clear();
    
    for (typename partition_t::block_iterator i = partition_->begin();
	 i != partition_->end(); i++)
      {
	hstate_t min = *i->begin();
	for (typename block_t::iterator j = i->begin(); j != i->end(); j++)
	  if (*j < min)
	    min = *j;
	states_.push_back(min);
      }
    
    sort(states_.begin(), states_.end());
    ////////////////////////////////////////////////////////////////////////////
    edges_.clear();
    
    for (typename automaton_t::edge_iterator i = auto_->edges().begin();
	 i != auto_->edges().end(); i++)
      {
	bool flag = true;
	for (edge_iterator j = edges_.begin(); j != edges_.end(); j++)
	  if ((origin_of(*i) == origin_of(*j)) &&
	      (aim_of(*i) == aim_of(*j)) &&
	      (label_of(*i) == label_of(*j)))
	    {
	      flag = false;
	      break;
	    }
	if (flag)
	  edges_.push_back(*i);
      }

    sort(edges_.begin(), edges_.end());
    ////////////////////////////////////////////////////////////////////////////
    initial_.clear();
    
    for (typename partition_t::block_iterator i = partition_->begin();
	 i != partition_->end(); i++)
      {
	bool flag = false;
	hstate_t min = *i->begin();
	for (typename block_t::iterator j = i->begin(); j != i->end(); j++)
	  {
	    if (auto_->is_initial(*j))
	      flag = true;
	    if (*j < min)
	      min = *j;
	  }
	if (flag)
	  initial_.push_back(min);
      }

    sort(initial_.begin(), initial_.end());
    ////////////////////////////////////////////////////////////////////////////
    final_.clear();
    for (typename partition_t::block_iterator i = partition_->begin();
	 i != partition_->end(); i++)
      {
	bool flag = false;
	hstate_t min = *i->begin();
	for (typename block_t::iterator j = i->begin(); j != i->end(); j++)
	  {
	    if (auto_->is_final(*j))
	      flag = true;
	    if (*j < min)
	      min = *j;
	  }
	if (flag)
	  final_.push_back(min);
      }
    
    sort(final_.begin(), final_.end());
  }

  template<typename T> 
  typename vector<hstate_t>::const_iterator
  Element<ViewLattice, T>::
  update(hstate_t state)
  {
    return find(states().begin(), states().end(), state);
  }

  template<typename T> 
  typename vector<hedge_t>::const_iterator
  Element<ViewLattice, T>::
  update(hedge_t edge)
  {
    return find(edges().begin(), edges().end(), edge);
  }
  
  template<typename T> 
  void 
  Element<ViewLattice, T>::
  update(state_iterator& lhs, state_iterator& rhs)
  {
    hstate_t lhs_state = *lhs;
    hstate_t rhs_state = *rhs;

    update();
    
    lhs = find(states().begin(), states().end(), lhs_state);
    rhs = find(states().begin(), states().end(), rhs_state);
  }

  template<typename T> 
  void 
  Element<ViewLattice, T>::
  update(edge_iterator& lhs, edge_iterator& rhs)
  {
    hedge_t lhs_edge = *lhs;
    hedge_t rhs_edge = *rhs;

    update();
    
    lhs = find(edges().begin(), edges().end(), lhs_edge);
    rhs = find(edges().begin(), edges().end(), rhs_edge);
  }

  template<typename T>   
  void 
  Element<ViewLattice, T>::
  update_init(initial_iterator& lhs, initial_iterator& rhs)
  {
    hstate_t lhs_state = *lhs;
    hstate_t rhs_state = *rhs;

    update();
    
    lhs = find(initial().begin(), initial().end(), lhs_state);
    rhs = find(initial().begin(), initial().end(), rhs_state);
  }

  template<typename T> 
  void 
  Element<ViewLattice, T>::
  update_final(final_iterator& lhs, final_iterator& rhs)
  {
    hstate_t lhs_state = *lhs;
    hstate_t rhs_state = *rhs;
    
    update();
    
    lhs = find(final().begin(), final().end(), lhs_state);
    rhs = find(final().begin(), final().end(), rhs_state);  
  }
    
  template<typename T> 
  const vector<hstate_t>&
  Element<ViewLattice, T>::states() const
  {
    return states_;
  }
  
  template<typename T> 
  const vector<hedge_t>&
  Element<ViewLattice, T>::edges() const
  {
    return edges_;
  }
  
  template<typename T> 
  const vector<hstate_t>&
  Element<ViewLattice, T>::initial() const
  {
    return initial_;
  }

  template<typename T> 
  const vector<hstate_t>&
  Element<ViewLattice, T>::final() const 
  {
    return final_;
  }
  
  template<typename T> 
  bool 
  Element<ViewLattice, T>::is_initial(hstate_t state) const
  {
    return (find(this->initial().begin(), this->initial().end(), state) != 
	    this->initial().end());
  }

  template<typename T> 
  bool 
  Element<ViewLattice, T>::is_final(hstate_t state) const
  {
    return (find(this->final().begin(), this->final().end(), state) != 
	    this->final().end());
  }

  template<typename T> 
  hstate_t
  Element<ViewLattice, T>::choose_state() const
  {
    return get_min_block(auto_->choose_state());
  }
  
  template<typename T> 
  bool
  Element<ViewLattice, T>::has_state(hstate_t s) const
  {
    vector<hstate_t> vector_ = states();

    return (find(vector_.begin(), vector_.end(), s) != vector_.end());
  }

  template<typename T> 
  bool
  Element<ViewLattice, T>::has_edge(hedge_t e) const
  {
    vector<hedge_t> vector_ = edges();
    
    return (find(vector_.begin(), vector_.end(), e) != vector_.end());
  }

  template<typename T> 
  hstate_t
  Element<ViewLattice, T>::origin_of(hedge_t e) const
  {
    return get_min_block(auto_->origin_of(e));
  }

  template<typename T> 
  hstate_t
  Element<ViewLattice, T>::aim_of(hedge_t e) const
  {
    return get_min_block(auto_->aim_of(e));
  }

  template<typename T> 
  typename automaton_traits<typename T::automaton_t::value_t>::label_ret_t
  Element<ViewLattice, T>::label_of(hedge_t e) const
  {
    return auto_->label_of(e);
  }

  template<typename T> 
  template<typename Container_>
  void 
  Element<ViewLattice, T>::
  deltac(Container_& res, hstate_t from, delta_kind::edges k) const
  {
    vector<hedge_t> tmp;
    
    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if (origin_of(*i) == from)
	tmp.push_back(*i);
    
    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  deltac(Container_& res, 
	 hstate_t from, 
	 const L& query,
	 delta_kind::edges k) const
  {
    vector<hedge_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((origin_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(*i);
    
    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_>
  void 
  Element<ViewLattice, T>::
  rdeltac(Container_& res, hstate_t from, delta_kind::edges k) const
  {
    vector<hedge_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if (aim_of(*i) == from)
	tmp.push_back(*i);

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  rdeltac(Container_& res, 
	  hstate_t from, 
	  const L& query,
	  delta_kind::edges k) const
  {
    vector<hedge_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((aim_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(*i);

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }
  
  template<typename T> 
  template<typename Container_>
  void 
  Element<ViewLattice, T>::
  deltac(Container_& res, hstate_t from, delta_kind::states k) const
  {
    vector<hstate_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if (origin_of(*i) == from)
	tmp.push_back(aim_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  deltac(Container_& res, 
	 hstate_t from, 
	 const L& query,
	 delta_kind::states k) const
  {
    vector<hstate_t> tmp;
    
    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((origin_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(aim_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_>
  void 
  Element<ViewLattice, T>::
  rdeltac(Container_& res, hstate_t from, delta_kind::states k) const
  {
    vector<hstate_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if (aim_of(*i) == from)
	tmp.push_back(origin_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  rdeltac(Container_& res, 
	  hstate_t from, 
	  const L& query,
	  delta_kind::states k) const
  {
    vector<hstate_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((aim_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(origin_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }



  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  letter_deltac(Container_& res, 
		hstate_t from, 
		const L& query,
		delta_kind::edges k) const
  {
    vector<hedge_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((origin_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(*i);
    
    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  letter_rdeltac(Container_& res, 
		 hstate_t from, 
		 const L& query,
		 delta_kind::edges k) const
  {
    vector<hedge_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((aim_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(*i);

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }
  
  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  letter_deltac(Container_& res, 
		hstate_t from, 
		const L& query,
		delta_kind::states k) const
  {
    vector<hstate_t> tmp;
    
    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((origin_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(aim_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }

  template<typename T> 
  template<typename Container_, typename L>	
  void 
  Element<ViewLattice, T>::
  letter_rdeltac(Container_& res, 
		 hstate_t from, 
		 const L& query,
		 delta_kind::states k) const
  {
    vector<hstate_t> tmp;

    for (edge_iterator i = edges_.begin(); i != edges_.end(); i++)
      if ((aim_of(*i) == from) && 
	  (*label_of(*i).begin()->first.begin() == query))
	tmp.push_back(origin_of(*i));

    if (tmp.size())
      copy(tmp.begin(), tmp.end(), inserter(res, res.begin()));
  }
  
  template<typename T> 
  typename automaton_traits<typename T::automaton_t::value_t>::series_t::monoid_t::alphabet_t&
  Element<ViewLattice, T>::
  alphabet() const
  {
    return auto_->series().monoid().alphabet();
  }

  template<typename T> 
  Element<ViewLattice, T>& 
  Element<ViewLattice, T>::
  operator = (const self_t& other)
  {
    states_ = other.states_;
    edges_ = other.edges_;
    initial_ = other.initial_;
    final_ = other.final_;
    
    auto_ = other.auto_;
    partition_ = other.partition_;
    
    return *this;
  }

} // mical

#endif // MICAL_VIEW_LATTICE_ELT_HXX_
