// depth_search.hxx
//
// Mical, a regular grammar inference library
// Copyright (C) 2003 Maxime Rey
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MICAL_DEPTH_SEARCH_HXX_
# define MICAL_DEPTH_SEARCH_HXX_

#include <mical/walkers/depth_search.hh>

#include <mical/tools/search_tools.hh>

namespace mical {

  template <class A, class T, class Task_, class Secure_>
  void
  depth_search(const Element<A, T>& a, 
	       Task_& task, 
	       security::level<Secure_> k)
  {
    Element<A, T> work = a;
    work.emancipate();

    depth_search(work, task, k);
  }
  
  template <class A, class T, class Task_, class Secure_>
  void
  depth_search(Element<A, T>& a, 
	       Task_& task, 
	       security::level<Secure_> k)
  {
    depth_search(a, *a.initial().begin(), task, k);
  }

  template <class A, class T, class Task_, class Secure_>
  void
  depth_search(const Element<A, T>& a, 
	       hstate_t init, 
	       Task_& task, 
	       security::level<Secure_> k)
  {
    Element<A, T> work = a;
    work.emancipate();
    
    depth_search(work, init, task, k);
  }

  template <class A, class T, class Task_, class Secure_>
  void
  depth_search(Element<A, T>& a, 
	       hstate_t init, 
	       Task_& task, 
	       security::level<Secure_> k)
  {
    typedef InfoDepth<A, T, Task_> info_t;

    info_t info(k.get_cycle_number());

    info.auto_ = a;
    info.auto_.emancipate();
    
    info.state_ = init;
    info.prev_state_ = init;
    info.edge_ = 0;
    info.task_ = task;

    for (typename Element<A, T>::state_iterator i = a.states().begin();
	 i != a.states().end(); i++)
      info.marked_[*i] = 0;

    depth_search_(info, Secure_());
    
    a = info.auto_;
    task = info.task_;
  }

  //////////////////////////////////////////////////////////////////////////////
  template <class A, class T, class Task_>
  void
  depth_search_(InfoDepth<A, T, Task_>& info, security::safe k)
  {
    if (info.stop_)
      return ;
    
    typedef typename Task_::delta_state_t            delta_state_t;
    typedef InfoDepth<A, T, Task_>          info_t;
    typedef typename info_t::marked_t       marked_t;

    info.task_.prefix(info);

  

    bool svg_first = info.first_;
    map<unsigned, unsigned> mappy;

    if (info.first_)
      info.first_ = false;
    else
      info.marked_[info.state_] += 1;

    hedge_t svg_edge = info.edge_;
    hstate_t svg_prev = info.prev_state_;
    hstate_t svg = info.state_;



    delta_state_t delta_state;
    info.task_.delta(info, delta_state, info.state_);

    give_map(delta_state, mappy);

    unsigned compt = 0;
    for (typename delta_state_t::iterator i = delta_state.begin();
	 i != delta_state.end(); i++)
      {

	unsigned pas = 0;
	for (typename marked_t::iterator j = info.marked_.begin(); 
	     j != info.marked_.end(); j++)
	  if (j->first == *i)
	    pas = j->second;
	
	if (pas < info.nb_cycles_)
	  {
	    info.prev_state_ = info.state_;
	    info.state_ = *i;
	    
	    set<hedge_t> delta_edge;
	    info.auto_.rdeltac(delta_edge, info.state_, delta_kind::edges());
	 
	    unsigned count = 0;
	    for (typename set<hedge_t>::iterator j = delta_edge.begin(); 
		 j != delta_edge.end(); j++)
	      if (info.auto_.origin_of(*j) == info.prev_state_)
		if (count < mappy[compt])
		  {
		    info.edge_ = *j; 
		    count++;
		  }

	    depth_search_(info, k);
	    
	    info.edge_ = svg_edge;
	    info.prev_state_ = svg_prev;    
	    info.state_ = svg;

	    if (++i != delta_state.end())
	      info.task_.infix(info);
	    --i;
	  }
	compt++;
      }
    info.first_ = svg_first;

    if (!info.first_)
      info.marked_[info.state_] -= 1;    
    
    info.task_.postfix(info);
  }
  //////////////////////////////////////////////////////////////////////////////

  //////////////////////////////////////////////////////////////////////////////
  template <class A, class T, class Task_>
  void
  depth_search_(InfoDepth<A, T, Task_>& info, security::responsable k)
  {
    if (info.stop_)
      return ;

    typedef typename Task_::delta_state_t            delta_state_t;

    info.task_.prefix(info);

    bool svg_first = info.first_;
    map<unsigned, unsigned> mappy;

    if (info.first_)
      info.first_ = false;
    
    delta_state_t delta_state;
    info.task_.delta(info, delta_state, info.state_);

    give_map(delta_state, mappy);

    hedge_t svg_edge = info.edge_;
    hstate_t svg_prev = info.prev_state_;
    hstate_t svg = info.state_;

    unsigned compt = 0;
    for (typename delta_state_t::iterator i = delta_state.begin();
	 i != delta_state.end(); i++)
      {
	info.prev_state_ = info.state_;
	info.state_ = *i;

	set<hedge_t> delta_edge;	
	info.auto_.rdeltac(delta_edge, info.state_, delta_kind::edges());
	
	unsigned count = 0;
	for (typename set<hedge_t>::iterator j = delta_edge.begin(); 
	     j != delta_edge.end(); j++)
	  if (info.auto_.origin_of(*j) == info.prev_state_)
	    if (count < mappy[compt])
	      {
		info.edge_ = *j; 
		count++;
	      }
	
	depth_search_(info, k);
	
	info.prev_state_ = svg_prev;    
	info.state_ = svg;
	info.edge_ = svg_edge;

	if (++i != delta_state.end())
	  info.task_.infix(info);
	--i;
	compt++;
      }

    info.first_ = svg_first;
    
    info.task_.postfix(info);
  }
  //////////////////////////////////////////////////////////////////////////////

} // mical

#endif // MICAL_DEPTH_SEARCH_HXX_
